import { RestApiProxy } from '../api/rest-api-proxy.js';
import { isEmptyObject } from '../common/utils.js';
import { EmptyTargetError, UniquenessError } from './errors.js';
import { mergeDefaultsWithPayload, pickUniqFieldsFromRecord, pickUniqFieldsFromRecords } from './utils.js';
export class Model extends RestApiProxy {
    constructor(modelName, schema, RushDBInstance) {
        super();
        this.label = modelName;
        this.schema = schema;
        RushDBInstance === null || RushDBInstance === void 0 ? void 0 : RushDBInstance.registerModel(this);
    }
    getLabel() {
        return this.label;
    }
    async find(params = {}, transaction) {
        var _a;
        return (_a = this.apiProxy) === null || _a === void 0 ? void 0 : _a.records.find(this.label, params, transaction);
    }
    async findOne(params = {}, transaction) {
        var _a;
        return (_a = this.apiProxy) === null || _a === void 0 ? void 0 : _a.records.findOne(this.label, Object.assign({}, params), transaction);
    }
    async findById(id, transaction) {
        return this.findOne({ where: { $id: id } }, transaction);
    }
    async findUniq(params = {}, transaction) {
        var _a;
        return (_a = this.apiProxy) === null || _a === void 0 ? void 0 : _a.records.findUniq(this.label, Object.assign({}, params), transaction);
    }
    async create(record, transaction) {
        var _a;
        const data = await mergeDefaultsWithPayload(this.schema, record);
        const uniqFields = pickUniqFieldsFromRecord(this.schema, data);
        if (!isEmptyObject(uniqFields)) {
            const tx = transaction !== null && transaction !== void 0 ? transaction : (await this.apiProxy.tx.begin());
            const matchingRecords = await this.find({ where: uniqFields }, tx);
            const hasOwnTransaction = typeof transaction !== 'undefined';
            const canCreate = !((_a = matchingRecords === null || matchingRecords === void 0 ? void 0 : matchingRecords.data) === null || _a === void 0 ? void 0 : _a.length);
            if (canCreate) {
                const result = await this.apiProxy.records.create(this.label, data, tx);
                if (!hasOwnTransaction) {
                    await tx.commit();
                }
                return result;
            }
            else {
                if (!hasOwnTransaction) {
                    await tx.commit();
                }
                // @TODO: Make it optional
                throw new UniquenessError(this.label, uniqFields);
            }
        }
        return await this.apiProxy.records.create(this.label, data, transaction);
    }
    attach(sourceId, target, options, transaction) {
        return this.apiProxy.records.attach(sourceId, target, options, transaction);
    }
    detach(sourceId, target, options, transaction) {
        return this.apiProxy.records.detach(sourceId, target, options, transaction);
    }
    async handleSetOrUpdate(id, record, method, transaction) {
        var _a;
        const data = await mergeDefaultsWithPayload(this.schema, record);
        const uniqFields = pickUniqFieldsFromRecord(this.schema, data);
        if (!isEmptyObject(uniqFields)) {
            const tx = transaction !== null && transaction !== void 0 ? transaction : (await this.apiProxy.tx.begin());
            const matchingRecords = await this.find({ where: uniqFields }, tx);
            const hasOwnTransaction = typeof transaction !== 'undefined';
            const canUpdate = !((_a = matchingRecords === null || matchingRecords === void 0 ? void 0 : matchingRecords.data) === null || _a === void 0 ? void 0 : _a.length) ||
                (matchingRecords.data.length === 1 && matchingRecords.data[0].__id === id);
            if (canUpdate) {
                const result = await this.apiProxy.records[method](id, data, tx);
                if (!hasOwnTransaction) {
                    await tx.commit();
                }
                return result;
            }
            else {
                if (!hasOwnTransaction) {
                    await tx.commit();
                }
                throw new UniquenessError(this.label, uniqFields);
            }
        }
        return await this.apiProxy.records[method](id, data, transaction);
    }
    async set(id, record, transaction) {
        return await this.handleSetOrUpdate(id, record, 'set', transaction);
    }
    async update(id, record, transaction) {
        return await this.handleSetOrUpdate(id, record, 'update', transaction);
    }
    async createMany(records, transaction) {
        var _a;
        // Begin a transaction if one isn't provided.
        const hasOwnTransaction = typeof transaction !== 'undefined';
        const tx = transaction !== null && transaction !== void 0 ? transaction : (await this.apiProxy.tx.begin());
        try {
            // Apply defaults in parallel
            const recordsToStore = await Promise.all(records.map(async (record) => {
                return await mergeDefaultsWithPayload(this.schema, record);
            }));
            // Check uniqueness
            const uniqProperties = pickUniqFieldsFromRecords(recordsToStore, this.schema, this.label);
            if (!isEmptyObject(uniqProperties)) {
                const criteria = Object.entries(uniqProperties).map(([key, values]) => ({
                    [key]: { $in: values }
                }));
                const matchingRecords = await this.find({
                    where: {
                        $or: criteria
                    }
                }, tx);
                if ((_a = matchingRecords === null || matchingRecords === void 0 ? void 0 : matchingRecords.data) === null || _a === void 0 ? void 0 : _a.length) {
                    throw new UniquenessError(this.label, Object.keys(uniqProperties));
                }
            }
            // Create records in the database
            const createdRecords = await this.apiProxy.records.createMany(this.label, recordsToStore, tx);
            // Commit the transaction if it was created internally
            if (!hasOwnTransaction) {
                await tx.commit();
            }
            return createdRecords;
        }
        catch (error) {
            if (!hasOwnTransaction) {
                await tx.rollback();
            }
            throw error;
        }
    }
    async delete(searchParams, transaction) {
        if (isEmptyObject(searchParams.where)) {
            throw new EmptyTargetError(`You must specify criteria to delete records of type '${this.label}'. Empty criteria are not allowed. If this was intentional, use the Dashboard instead.`);
        }
        return await this.apiProxy.records.delete(Object.assign(Object.assign({}, searchParams), { labels: [this.label] }), transaction);
    }
    async deleteById(idOrIds, transaction) {
        return await this.apiProxy.records.deleteById(idOrIds, transaction);
    }
}
